# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

import typer

from ..logging import TracebackMarkupConsole
from ..version import __version__

from .auth import app as auth_app
from .compile import compile_predictor
from .misc import cli_options
from .predictions import create_prediction
from .predictors import archive_predictor, delete_predictor, retrieve_predictor
from .sources import retrieve_source

# Define CLI
typer.main.console_stderr = TracebackMarkupConsole()
app = typer.Typer(
    name=f"Function CLI {__version__}",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
    add_completion=False
)

# Add top level options
app.callback()(cli_options)

# Add subcommands
app.add_typer(auth_app, name="auth", help="Login, logout, and check your authentication status.")

# Add top-level commands
app.command(
    name="predict",
    help="Make a prediction.",
    context_settings={ "allow_extra_args": True, "ignore_unknown_options": True }
)(create_prediction)
app.command(
    name="compile",
    help="Create a predictor by compiling a Python function."
)(compile_predictor)
app.command(name="retrieve", help="Retrieve a predictor.")(retrieve_predictor)
app.command(name="archive", help="Archive a predictor.")(archive_predictor)
app.command(name="delete", help="Delete a predictor.")(delete_predictor)
app.command(name="source", help="Retrieve the native source code for a given prediction.")(retrieve_source)

# Run
if __name__ == "__main__":
    app()


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import CDLL
from enum import IntEnum
from importlib import resources
from platform import machine, system

_fxnc: CDLL = None

class FXNStatus(IntEnum):
    OK = 0
    ERROR_INVALID_ARGUMENT = 1
    ERROR_INVALID_OPERATION = 2
    ERROR_NOT_IMPLEMENTED = 3

def get_fxnc () -> CDLL:
    global _fxnc
    _fxnc = _fxnc if _fxnc is not None else _load_fxnc()
    return _fxnc

def set_fxnc (fxnc: CDLL):
    global _fxnc
    _fxnc = fxnc

def _load_fxnc () -> CDLL:
    os = system().lower()
    os = "macos" if os == "darwin" else os
    arch = machine().lower()
    arch = "arm64" if arch == "aarch64" else arch
    arch = "x86_64" if arch in ["x64", "amd64"] else arch
    package = f"fxn.lib.{os}.{arch}"
    resource = "libFunction.so"
    resource = "Function.dylib" if os == "macos" else resource
    resource = "Function.dll" if os == "windows" else resource
    with resources.path(package, resource) as path:
        return CDLL(str(path))
    
def status_to_error (status: int) -> str:
    if status == FXNStatus.ERROR_INVALID_ARGUMENT:
        return "FXN_ERROR_INVALID_ARGUMENT"
    elif status == FXNStatus.ERROR_INVALID_OPERATION:
        return "FXN_ERROR_INVALID_OPERATION"
    elif status == FXNStatus.ERROR_NOT_IMPLEMENTED:
        return "FXN_ERROR_NOT_IMPLEMENTED"
    return ""


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ..client import FunctionClient
from .remote import RemotePredictionService

class PredictionService:
    """
    Make predictions.
    """
    remote: RemotePredictionService

    def __init__ (self, client: FunctionClient):
        self.remote = RemotePredictionService(client)


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from asyncio import run as run_async
from importlib.util import module_from_spec, spec_from_file_location
from inspect import getmembers, getmodulename, isfunction
from pathlib import Path
from pydantic import BaseModel
from rich import print as print_rich
import sys
from typer import Argument, Option
from typing import Callable, Literal
from urllib.parse import urlparse, urlunparse

from ..client import FunctionAPIError
from ..compile import PredictorSpec
from ..function import Function
from ..sandbox import EntrypointCommand
from ..logging import CustomProgress, CustomProgressTask
from .auth import get_access_key

class CompileError (Exception):
    pass

def compile_predictor (
    path: str=Argument(..., help="Predictor path."),
    overwrite: bool=Option(False, "--overwrite", help="Whether to delete any existing predictor with the same tag before compiling."),
):
    run_async(_compile_predictor_async(path, overwrite=overwrite))

async def _compile_predictor_async (
    path: str,
    *,
    overwrite: bool
):
    fxn = Function(get_access_key())
    path: Path = Path(path).resolve()
    with CustomProgress():
        # Load
        with CustomProgressTask(loading_text="Loading predictor...") as task:
            func = _load_predictor_func(path)
            entrypoint = EntrypointCommand(from_path=str(path), to_path="./", name=func.__name__)
            spec: PredictorSpec = func.__predictor_spec
            task.finish(f"Loaded prediction function: [bold cyan]{spec.tag}[/bold cyan]")
        # Populate
        sandbox = spec.sandbox
        sandbox.commands.append(entrypoint)
        with CustomProgressTask(loading_text="Uploading sandbox...", done_text="Uploaded sandbox"):
            sandbox.populate(fxn=fxn)
        # Compile
        with CustomProgressTask(loading_text="Running codegen...", done_text="Completed codegen"):
            with CustomProgressTask(loading_text="Creating predictor..."):
                if overwrite:
                    try:
                        fxn.client.request(
                            method="DELETE",
                            path=f"/predictors/{spec.tag}"
                        )
                    except FunctionAPIError as error:
                        if error.status_code != 404:
                            raise
                predictor = fxn.client.request(
                    method="POST",
                    path="/predictors",
                    body=spec.model_dump(mode="json", exclude=spec.model_extra.keys(), by_alias=True),
                    response_type=_Predictor
                )
            with ProgressLogQueue() as task_queue:
                async for event in fxn.client.stream(
                    method="POST",
                    path=f"/predictors/{predictor.tag}/compile",
                    body={ },
                    response_type=_LogEvent | _ErrorEvent
                ):
                    if isinstance(event, _LogEvent):
                        task_queue.push_log(event)
                    elif isinstance(event, _ErrorEvent):
                        task_queue.push_error(event)
                        raise CompileError(event.data.error)
    predictor_url = _compute_predictor_url(fxn.client.api_url, spec.tag)
    print_rich(f"\n[bold spring_green3]🎉 Predictor is now being compiled.[/bold spring_green3] Check it out at [link={predictor_url}]{predictor_url}[/link]")

def _load_predictor_func (path: str) -> Callable[...,object]:
    if "" not in sys.path:
        sys.path.insert(0, "")
    path: Path = Path(path).resolve()
    sys.path.insert(0, str(path.parent))
    name = getmodulename(path)
    spec = spec_from_file_location(name, path)
    module = module_from_spec(spec)
    sys.modules[name] = module
    spec.loader.exec_module(module)
    main_func = next(func for _, func in getmembers(module, isfunction) if hasattr(func, "__predictor_spec"))
    return main_func

def _compute_predictor_url (api_url: str, tag: str) -> str:
    parsed_url = urlparse(api_url)
    hostname_parts = parsed_url.hostname.split(".")
    if hostname_parts[0] == "api":
        hostname_parts.pop(0)
    hostname = ".".join(hostname_parts)
    netloc = hostname if not parsed_url.port else f"{hostname}:{parsed_url.port}"
    predictor_url = urlunparse(parsed_url._replace(netloc=netloc, path=f"{tag}"))
    return predictor_url

class _Predictor (BaseModel):
    tag: str

class _LogData (BaseModel):
    message: str
    level: int = 0
    status: Literal["success", "error"] = "success"
    update: bool = False

class _LogEvent (BaseModel):
    event: Literal["log"]
    data: _LogData

class _ErrorData (BaseModel):
    error: str

class _ErrorEvent (BaseModel):
    event: Literal["error"]
    data: _ErrorData

class ProgressLogQueue:

    def __init__ (self):
        self.queue: list[tuple[int, CustomProgressTask]] = []

    def push_log (self, event: _LogEvent):
        # Check for update
        if event.data.update and self.queue:
            current_level, current_task = self.queue[-1]
            current_task.update(description=event.data.message, status=event.data.status)
            return
        # Pop
        while self.queue:
            current_level, current_task = self.queue[-1]
            if event.data.level > current_level:
                break
            current_task.__exit__(None, None, None)
            self.queue.pop()
        task = CustomProgressTask(loading_text=event.data.message)
        task.__enter__()
        self.queue.append((event.data.level, task))

    def push_error (self, error: _ErrorEvent):
        while self.queue:
            _, current_task = self.queue.pop()
            current_task.__exit__(RuntimeError, None, None)

    def __enter__ (self):
        return self
    
    def __exit__ (self, exc_type, exc_value, traceback):
        while self.queue:
            _, current_task = self.queue.pop()
            current_task.__exit__(None, None, None)


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ..client import FunctionClient, FunctionAPIError
from ..types import Predictor

class PredictorService:

    def __init__ (self, client: FunctionClient) -> None:
        self.client = client

    def retrieve (self, tag: str) -> Predictor:
        """
        Retrieve a predictor.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            Predictor: Predictor.
        """
        try:
            return self.client.request(
                method="GET",
                path=f"/predictors/{tag}",
                response_type=Predictor
            )
        except FunctionAPIError as error:
            if error.status_code == 404:
                return None
            raise


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from rich import print_json
from typer import Argument

from ..function import Function
from ..logging import CustomProgress, CustomProgressTask
from .auth import get_access_key

def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    with CustomProgress(transient=True):
        with CustomProgressTask(loading_text="Retrieving predictor..."):
            fxn = Function(get_access_key())
            predictor = fxn.predictors.retrieve(tag)
            predictor = predictor.model_dump() if predictor else None
            print_json(data=predictor)

def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    with CustomProgress():
        with CustomProgressTask(
            loading_text="Archiving predictor...",
            done_text=f"Archived predictor: [bold dark_orange]{tag}[/bold dark_orange]"
        ):
            fxn = Function(get_access_key())
            fxn.client.request(
                method="POST",
                path=f"/predictors/{tag}/archive"
            )

def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
):
    with CustomProgress():
        with CustomProgressTask(
            loading_text="Deleting predictor...",
            done_text=f"Deleted predictor: [bold red]{tag}[/bold red]"
        ):
            fxn = Function(get_access_key())
            fxn.client.request(
                method="DELETE",
                path=f"/predictors/{tag}"
            )


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from json import loads, JSONDecodeError
from pydantic import BaseModel, TypeAdapter
from requests import request
from typing import AsyncGenerator, Literal, Type, TypeVar

T = TypeVar("T", bound=BaseModel)

class FunctionClient:
    
    def __init__(self, access_key: str, api_url: str | None) -> None:
        self.access_key = access_key
        self.api_url = api_url or "https://api.fxn.ai/v1"

    def request (
        self,
        *,
        method: Literal["GET", "POST", "PATCH", "DELETE"],
        path: str,
        body: dict[str, object]=None,
        response_type: Type[T]=None
    ) -> T:
        """
        Make a request to a REST endpoint.

        Parameters:
            method (str): Request method.
            path (str): Endpoint path.
            body (dict): Request JSON body.
            response_type (Type): Response type.
        """
        response = request(
            method=method,
            url=f"{self.api_url}{path}",
            json=body,
            headers={ "Authorization": f"Bearer {self.access_key}" }
        )
        data = response.text
        try:
            data = response.json()
        except JSONDecodeError:
            pass
        if response.ok:
            return response_type(**data) if response_type is not None else None
        else:
            error = _ErrorResponse(**data).errors[0].message if isinstance(data, dict) else data
            raise FunctionAPIError(error, response.status_code)
        
    async def stream (
        self,
        *,
        method: Literal["GET", "POST", "PATCH", "DELETE"],
        path: str,
        body: dict[str, object]=None,
        response_type: Type[T]=None
    ) -> AsyncGenerator[T, None]:
        """
        Make a request to a REST endpoint and consume the response as a server-sent events stream.

        Parameters:
            method (str): Request method.
            path (str): Endpoint path.
            body (dict): Request JSON body.
            response_type (Type): Response type.
        """
        response = request(
            method=method,
            url=f"{self.api_url}{path}",
            json=body,
            headers={
                "Accept": "text/event-stream",
                "Authorization": f"Bearer {self.access_key}"
            },
            stream=True
        )
        event = None
        data: str = ""
        for line in response.iter_lines(decode_unicode=True):
            if line is None:
                break
            line: str = line.strip()
            if line:
                if line.startswith("event:"):
                    event = line[len("event:"):].strip()
                elif line.startswith("data:"):
                    line_data = line[len("data:"):].strip()
                    data = f"{data}\n{line_data}"
                continue
            if event is not None:
                yield _parse_sse_event(event, data, response_type)
            event = None
            data = ""
        if event or data:
            yield _parse_sse_event(event, data, response_type)

class FunctionAPIError (Exception):

    def __init__(self, message: str, status_code: int):
        super().__init__(message)
        self.message = message
        self.status_code = status_code

    def __str__(self):
        return f"FunctionAPIError: {self.message} (Status Code: {self.status_code})"
    
class _APIError (BaseModel):
    message: str

class _ErrorResponse (BaseModel):
    errors: list[_APIError]

def _parse_sse_event (event: str, data: str, type: Type[T]=None) -> T:
    result = { "event": event, "data": loads(data) }
    result = TypeAdapter(type).validate_python(result) if type is not None else result
    return result


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import byref, c_int, c_int32, c_void_p, create_string_buffer
from pathlib import Path
from typing import final

from .fxnc import get_fxnc, status_to_error, FXNStatus
from .value import Value

@final
class ValueMap:

    def __init__ (self, map=None, *, owner: bool=True):
        if map is None:
            map = c_void_p()
            owner = True
            status = get_fxnc().FXNValueMapCreate(byref(map))
            if status != FXNStatus.OK:
                raise RuntimeError(f"Failed to create value map with error: {status_to_error(status)}")
        self.__map = map
        self.__owner = owner

    def key (self, index: int) -> str:
        buffer = create_string_buffer(256)
        status = get_fxnc().FXNValueMapGetKey(self.__map, index, buffer, len(buffer))
        if status == FXNStatus.OK:
            return buffer.value.decode("utf-8")
        else:
            raise RuntimeError(f"Failed to get value map key at index {index} with error: {status_to_error(status)}")

    def __getitem__ (self, key: str) -> Value | None:
        value = c_void_p()
        status = get_fxnc().FXNValueMapGetValue(self.__map, key.encode(), byref(value))
        if status == FXNStatus.OK:
            return Value(value, owner=False)
        else:
            raise RuntimeError(f"Failed to get value map value for key '{key}' with error: {status_to_error(status)}")

    def __setitem__ (self, key: str, value: Value):
        status = get_fxnc().FXNValueMapSetValue(self.__map, key.encode(), value._Value__value)
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to set value map value for key '{key}' with error: {status_to_error(status)}")

    def __len__ (self) -> int:
        count = c_int32()
        status = get_fxnc().FXNValueMapGetSize(self.__map, byref(count))
        if status == FXNStatus.OK:
            return count.value
        else:
            raise RuntimeError(f"Failed to get value map size with error: {status_to_error(status)}")

    def __enter__ (self):
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__map and self.__owner:
            get_fxnc().FXNValueMapRelease(self.__map)
        self.__map = None


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from .client import FunctionAPIError
from .compile import compile, CompileMetadata, CompileTarget
from .function import Function
from .sandbox import Sandbox
from .types import *
from .version import __version__


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from pathlib import Path
from pydantic import BaseModel, BeforeValidator, ConfigDict, Field
from typing import Annotated, Literal

def _validate_torch_module (module: "torch.nn.Module") -> "torch.nn.Module": # type: ignore
    try:
        from torch.nn import Module # type: ignore
        if not isinstance(module, Module):
            raise ValueError(f"Expected torch.nn.Module, got {type(module)}")
        return module
    except ImportError:
        raise ImportError("PyTorch is required to create this metadata but is not installed.")

def _validate_ort_inference_session (session: "onnxruntime.InferenceSession") -> "onnxruntime.InferenceSession": # type: ignore
    try:
        from onnxruntime import InferenceSession # type: ignore
        if not isinstance(session, InferenceSession):
            raise ValueError(f"Expected onnxruntime.InferenceSession, got {type(session)}")
        return session
    except ImportError:
        raise ImportError("ONNXRuntime is required to create this metadata but is not installed.")

class CoreMLInferenceMetadata (BaseModel):
    """
    Metadata required to lower a PyTorch model for inference on iOS, macOS, and visionOS with CoreML.
    """
    kind: Literal["meta.inference.coreml"] = "meta.inference.coreml"
    model: Annotated[object, BeforeValidator(_validate_torch_module)] = Field(description="PyTorch module to apply metadata to.")
    model_args: list[object] = Field(description="Positional inputs to the model.")
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

class ONNXInferenceMetadata (BaseModel):
    """
    Metadata required to lower a PyTorch model for inference.
    """
    kind: Literal["meta.inference.onnx"] = "meta.inference.onnx"
    model: Annotated[object, BeforeValidator(_validate_torch_module)] = Field(description="PyTorch module to apply metadata to.")
    model_args: list[object] = Field(description="Positional inputs to the model.")
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

class ONNXRuntimeInferenceSessionMetadata (BaseModel):
    """
    Metadata required to lower an ONNXRuntime `InferenceSession` for inference.
    """
    kind: Literal["meta.inference.onnxruntime"] = "meta.inference.onnxruntime"
    session: Annotated[object, BeforeValidator(_validate_ort_inference_session)] = Field(description="ONNXRuntime inference session to apply metadata to.")
    model_path: Path = Field(description="ONNX model path. The model must exist at this path in the compiler sandbox.")
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

class LiteRTInferenceMetadata (BaseModel):
    """
    Metadata required to lower PyTorch model for inference with LiteRT (fka TensorFlow Lite).
    """
    kind: Literal["meta.inference.litert"] = "meta.inference.litert"
    model: Annotated[object, BeforeValidator(_validate_torch_module)] = Field(description="PyTorch module to apply metadata to.")
    model_args: list[object] = Field(description="Positional inputs to the model.")
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)

class LlamaCppInferenceMetadata (BaseModel): # INCOMPLETE
    """
    Metadata required to lower a GGUF model for LLM inference.
    """
    kind: Literal["meta.inference.gguf"] = "meta.inference.gguf"
    model_path: Path = Field(description="GGUF model path. The model must exist at this path in the compiler sandbox.")
    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=True)


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from collections.abc import Iterable
from enum import IntFlag
from ctypes import byref, cast, c_char_p, c_int, c_int32, c_uint8, c_void_p, string_at, POINTER
from io import BytesIO
from json import dumps, loads
from numpy import array, dtype, int32, ndarray, zeros
from numpy.ctypeslib import as_array, as_ctypes_type
from PIL import Image
from typing import final, Any

from ..types import Dtype
from .fxnc import get_fxnc, status_to_error, FXNStatus

class ValueFlags (IntFlag):
    NONE = 0
    COPY_DATA = 1

@final
class Value:

    def __init__ (self, value, *, owner: bool=True):
        self.__value = value
        self.__owner = owner

    @property
    def data (self):
        data = c_void_p()
        status = get_fxnc().FXNValueGetData(self.__value, byref(data))
        if status == FXNStatus.OK:
            return data
        else:
            raise RuntimeError(f"Failed to get value data with error: {status_to_error(status)}")

    @property
    def type (self) -> Dtype:
        dtype = c_int()
        status = get_fxnc().FXNValueGetType(self.__value, byref(dtype))
        if status == FXNStatus.OK:
            return _DTYPE_TO_STR.get(dtype.value)
        else:
            raise RuntimeError(f"Failed to get value data type with error: {status_to_error(status)}")

    @property
    def shape (self) -> list[int] | None:
        if self.type not in _TENSOR_ISH_DTYPES:
            return None
        fxnc = get_fxnc()
        dims = c_int32()
        status = fxnc.FXNValueGetDimensions(self.__value, byref(dims))
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get value dimensions with error: {status_to_error(status)}")
        shape = zeros(dims.value, dtype=int32)
        status = fxnc.FXNValueGetShape(self.__value, shape.ctypes.data_as(POINTER(c_int32)), dims)
        if status == FXNStatus.OK:
            return shape.tolist()
        else:
            raise RuntimeError(f"Failed to get value shape with error: {status_to_error(status)}")

    def to_object (self) -> Any:
        type = self.type
        if type == Dtype.null:
            return None
        elif type in _TENSOR_DTYPES:
            ctype = as_ctypes_type(dtype(type))
            tensor = as_array(cast(self.data, POINTER(ctype)), self.shape)
            return tensor.item() if len(tensor.shape) == 0 else tensor.copy()
        elif type == Dtype.string:
            return cast(self.data, c_char_p).value.decode()
        elif type in [Dtype.list, Dtype.dict]:
            return loads(cast(self.data, c_char_p).value.decode())
        elif type == Dtype.image:
            pixel_buffer = as_array(cast(self.data, POINTER(c_uint8)), self.shape)
            return Image.fromarray(pixel_buffer.squeeze()).copy()
        elif type == Dtype.binary:
            return BytesIO(string_at(self.data, self.shape[0]))
        else:
            raise RuntimeError(f"Failed to convert Function value to object because value has unsupported type: {type}")

    def __enter__ (self):
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__value and self.__owner:
            get_fxnc().FXNValueRelease(self.__value)
        self.__value = None

    @classmethod
    def create_array (
        cls,
        data: ndarray,
        *,
        flags: ValueFlags=ValueFlags.NONE
    ) -> Value:
        dtype = _STR_TO_DTYPE.get(data.dtype.name)
        if dtype is None:
            raise RuntimeError(f"Failed to create array value because data type is not supported: {data.dtype}")
        value = c_void_p()
        status = get_fxnc().FXNValueCreateArray(
            data.ctypes.data_as(c_void_p),
            data.ctypes.shape_as(c_int32),
            len(data.shape),
            dtype,
            flags,
            byref(value)
        )
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create array value with error: {status_to_error(status)}")

    @classmethod
    def create_string (cls, data: str) -> Value:
        value = c_void_p()
        status = get_fxnc().FXNValueCreateString(data.encode(), byref(value))
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create string value with error: {status_to_error(status)}")

    @classmethod
    def create_list (cls, data: Iterable[Any]) -> Value:
        value = c_void_p()
        status = get_fxnc().FXNValueCreateList(dumps(data).encode(), byref(value))
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create list value with error: {status_to_error(status)}")
    
    @classmethod
    def create_dict (cls, data: dict[str, Any]) -> Value:
        value = c_void_p()
        status = get_fxnc().FXNValueCreateDict(dumps(data).encode(), byref(value))
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create dict value with error: {status_to_error(status)}")

    @classmethod
    def create_image (cls, image: Image.Image) -> Value:
        value = c_void_p()
        pixel_buffer = array(image)
        status = get_fxnc().FXNValueCreateImage(
            pixel_buffer.ctypes.data_as(c_void_p),
            image.width,
            image.height,
            pixel_buffer.shape[2],
            ValueFlags.COPY_DATA,
            byref(value)
        )
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create image value with error: {status_to_error(status)}")        

    @classmethod
    def create_binary (
        cls,
        data: memoryview,
        *,
        flags: ValueFlags=ValueFlags.NONE
    ) -> Value:
        buffer = (c_uint8 * len(data)).from_buffer(data)
        value = c_void_p()
        status = get_fxnc().FXNValueCreateBinary(buffer, len(data), flags, byref(value))
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create binary value with error: {status_to_error(status)}")

    @classmethod
    def create_null (cls) -> Value:
        value = c_void_p()
        status = get_fxnc().FXNValueCreateNull(byref(value))
        if status == FXNStatus.OK:
            return Value(value)
        else:
            raise RuntimeError(f"Failed to create null value with error: {status_to_error(status)}")


_STR_TO_DTYPE = {
    Dtype.null: 0,
    Dtype.float16: 1,
    Dtype.float32: 2,
    Dtype.float64: 3,
    Dtype.int8: 4,
    Dtype.int16: 5,
    Dtype.int32: 6,
    Dtype.int64: 7,
    Dtype.uint8: 8,
    Dtype.uint16: 9,
    Dtype.uint32: 10,
    Dtype.uint64: 11,
    Dtype.bool: 12,
    Dtype.string: 13,
    Dtype.list: 14,
    Dtype.dict: 15,
    Dtype.image: 16,
    Dtype.binary: 17,
}
_DTYPE_TO_STR = { value: key for key, value in _STR_TO_DTYPE.items() }
_TENSOR_DTYPES = {
    Dtype.float16,
    Dtype.float32,
    Dtype.float64,
    Dtype.int8,
    Dtype.int16,
    Dtype.int32,
    Dtype.int64,
    Dtype.uint8,
    Dtype.uint16,
    Dtype.uint32,
    Dtype.uint64,
    Dtype.bool,
}
_TENSOR_ISH_DTYPES = _TENSOR_DTYPES | { Dtype.image }


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from base64 import b64encode
from dataclasses import asdict, is_dataclass
from enum import Enum
from io import BytesIO
from json import dumps, loads
from numpy import array, frombuffer, ndarray
from PIL import Image
from pydantic import BaseModel, Field
from requests import get, put
from typing import Literal
from urllib.request import urlopen

from ..c import Configuration
from ..client import FunctionClient
from ..services import Value
from ..types import Dtype, Prediction

RemoteAcceleration = Literal["auto", "cpu", "a40", "a100"]

class RemotePredictionService:
    """
    Make remote predictions.
    """

    def __init__ (self, client: FunctionClient):
        self.client = client

    def create (
        self,
        tag: str,
        *,
        inputs: dict[str, Value],
        acceleration: RemoteAcceleration="auto"
    ) -> Prediction:
        """
        Create a remote prediction.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values.
            acceleration (RemoteAcceleration): Prediction acceleration.

        Returns:
            Prediction: Created prediction.
        """
        input_map = { name: self.__to_value(value, name=name).model_dump(mode="json") for name, value in inputs.items() }
        prediction = self.client.request(
            method="POST",
            path="/predictions/remote",
            body={
                "tag": tag,
                "inputs": input_map,
                "acceleration": acceleration,
                "clientId": Configuration.get_client_id()
            },
            response_type=RemotePrediction
        )
        results = list(map(self.__to_object, prediction.results)) if prediction.results is not None else None
        prediction = Prediction(**{ **prediction.model_dump(), "results": results })
        return prediction

    def __to_value (
        self,
        obj: Value,
        *,
        name: str,
        max_data_url_size: int=4 * 1024 * 1024
    ) -> RemoteValue:
        obj = self.__try_ensure_serializable(obj)
        if obj is None:
            return RemoteValue(data=None, type=Dtype.null)
        elif isinstance(obj, float):
            obj = array(obj, dtype=Dtype.float32)
            return self.__to_value(obj, name=name, max_data_url_size=max_data_url_size)
        elif isinstance(obj, bool):
            obj = array(obj, dtype=Dtype.bool)
            return self.__to_value(obj, name=name, max_data_url_size=max_data_url_size)
        elif isinstance(obj, int):
            obj = array(obj, dtype=Dtype.int32)
            return self.__to_value(obj, name=name, max_data_url_size=max_data_url_size)
        elif isinstance(obj, ndarray):
            buffer = BytesIO(obj.tobytes())
            data = self.__upload(buffer, name=name, max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=obj.dtype.name, shape=list(obj.shape))
        elif isinstance(obj, str):
            buffer = BytesIO(obj.encode())
            data = self.__upload(buffer, name=name, mime="text/plain", max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=Dtype.string)
        elif isinstance(obj, list):
            buffer = BytesIO(dumps(obj).encode())
            data = self.__upload(buffer, name=name, mime="application/json", max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=Dtype.list)
        elif isinstance(obj, dict):
            buffer = BytesIO(dumps(obj).encode())
            data = self.__upload(buffer, name=name, mime="application/json", max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=Dtype.dict)
        elif isinstance(obj, Image.Image):
            buffer = BytesIO()
            format = "PNG" if obj.mode == "RGBA" else "JPEG"
            mime = f"image/{format.lower()}"
            obj.save(buffer, format=format)
            data = self.__upload(buffer, name=name, mime=mime, max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=Dtype.image)
        elif isinstance(obj, BytesIO):
            data = self.__upload(obj, name=name, max_data_url_size=max_data_url_size)
            return RemoteValue(data=data, type=Dtype.binary)
        else:
            raise ValueError(f"Failed to serialize value '{obj}' of type `{type(obj)}` because it is not supported")

    def __to_object (self, value: RemoteValue) -> Value:
        if value.type == Dtype.null:
            return None
        buffer = self.__download(value.data)
        if value.type in [
            Dtype.int8, Dtype.int16, Dtype.int32, Dtype.int64,
            Dtype.uint8, Dtype.uint16, Dtype.uint32, Dtype.uint64,
            Dtype.float16, Dtype.float32, Dtype.float64, Dtype.bool
        ]:
            assert value.shape is not None, "Array value must have a shape specified"
            array = frombuffer(buffer.getbuffer(), dtype=value.type).reshape(value.shape)
            return array if len(value.shape) > 0 else array.item()
        elif value.type == Dtype.string:
            return buffer.getvalue().decode("utf-8")
        elif value.type in [Dtype.list, Dtype.dict]:
            return loads(buffer.getvalue().decode("utf-8"))
        elif value.type == Dtype.image:
            return Image.open(buffer)
        elif value.type == Dtype.binary:
            return buffer
        else:
            raise ValueError(f"Failed to deserialize value with type `{value.type}` because it is not supported")

    def __upload (
        self,
        data: BytesIO,
        *,
        name: str,
        mime: str="application/octet-stream",
        max_data_url_size: int=4 * 1024 * 1024
    ) -> str:
        if data.getbuffer().nbytes <= max_data_url_size:
            encoded_data = b64encode(data.getvalue()).decode("ascii")
            return f"data:{mime};base64,{encoded_data}"
        value = self.client.request(
            method="POST",
            path="/values",
            body={ "name": name },
            response_type=CreateValueResponse
        )
        put(
            value.upload_url,
            data=data,
            headers={ "Content-Type": mime }
        ).raise_for_status()
        return value.download_url

    def __download (self, url: str) -> BytesIO:
        if url.startswith("data:"):
            with urlopen(url) as response:
                return BytesIO(response.read())
        response = get(url)
        response.raise_for_status()
        result = BytesIO(response.content)
        return result

    @classmethod
    def __try_ensure_serializable (cls, obj: object) -> object:
        if obj is None:
            return obj
        if isinstance(obj, list):
            return [cls.__try_ensure_serializable(x) for x in obj]
        if is_dataclass(obj) and not isinstance(obj, type):
            return asdict(obj)
        if isinstance(obj, BaseModel):
            return obj.model_dump(mode="json", by_alias=True)
        return obj

class RemoteValue (BaseModel):
    data: str | None
    type: Dtype
    shape: list[int] | None = None

class RemotePrediction (BaseModel):
    id: str
    tag: str
    created: str
    results: list[RemoteValue] | None
    latency: float | None
    error: str | None
    logs: str | None

class CreateValueResponse (BaseModel):
    upload_url: str = Field(validation_alias="uploadUrl")
    download_url: str = Field(validation_alias="downloadUrl")


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import byref, c_double, c_int32, c_void_p, create_string_buffer
from pathlib import Path
from typing import final

from .fxnc import get_fxnc, status_to_error, FXNStatus
from .map import ValueMap

@final
class Prediction:

    def __init__ (self, prediction):
        self.__prediction = prediction

    @property
    def id (self) -> str:
        id = create_string_buffer(256)
        status = get_fxnc().FXNPredictionGetID(self.__prediction, id, len(id))
        if status == FXNStatus.OK:
            return id.value.decode("utf-8")
        else:
            raise RuntimeError(f"Failed to get prediction id with error: {status_to_error(status)}")
    
    @property
    def latency (self) -> float:
        latency = c_double()
        status = get_fxnc().FXNPredictionGetLatency(self.__prediction, byref(latency))
        if status == FXNStatus.OK:
            return latency.value
        else:
            raise RuntimeError(f"Failed to get prediction latency with error: {status_to_error(status)}")

    @property
    def results (self) -> ValueMap | None:
        map = c_void_p()
        status = get_fxnc().FXNPredictionGetResults(self.__prediction, byref(map))
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get prediction results with error: {status_to_error(status)}")
        map = ValueMap(map, owner=False)
        return map if len(map) > 0 else None

    @property
    def error (self) -> str | None:
        error = create_string_buffer(2048)
        get_fxnc().FXNPredictionGetError(self.__prediction, error, len(error))
        error = error.value.decode("utf-8")
        return error if error else None

    @property
    def logs (self) -> str:
        fxnc = get_fxnc()
        log_length = c_int32()
        status = fxnc.FXNPredictionGetLogLength(self.__prediction, byref(log_length))
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get prediction log length with error: {status_to_error(status)}")
        logs = create_string_buffer(log_length.value + 1)
        status = fxnc.FXNPredictionGetLogs(self.__prediction, logs, len(logs))
        if status == FXNStatus.OK:
            return logs.value.decode("utf-8")
        else:
            raise RuntimeError(f"Failed to get prediction logs with error: {status_to_error(status)}")

    def __enter__ (self):
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__prediction:
            get_fxnc().FXNPredictionRelease(self.__prediction)
        self.__prediction = None


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import byref, c_int, c_void_p, create_string_buffer
from pathlib import Path
from typing import final

from ..types import Acceleration
from .fxnc import get_fxnc, status_to_error, FXNStatus

@final
class Configuration:

    def __init__ (self):
        configuration = c_void_p()
        status = get_fxnc().FXNConfigurationCreate(byref(configuration))
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to create configuration with error: {status_to_error(status)}")
        self.__configuration = configuration            

    @property
    def tag (self) -> str:
        buffer = create_string_buffer(2048)
        status = get_fxnc().FXNConfigurationGetTag(
            self.__configuration,
            buffer,
            len(buffer)
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get configuration tag with error: {status_to_error(status)}")
        tag = buffer.value.decode("utf-8")
        return tag if tag else None

    @tag.setter
    def tag (self, tag: str):
        tag = tag.encode() if tag is not None else None
        status = get_fxnc().FXNConfigurationSetTag(self.__configuration, tag)
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to set configuration tag with error: {status_to_error(status)}")

    @property
    def token (self) -> str:
        buffer = create_string_buffer(2048)
        status = get_fxnc().FXNConfigurationGetToken(
            self.__configuration,
            buffer,
            len(buffer)
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get configuration token with error: {status_to_error(status)}")
        token = buffer.value.decode("utf-8")
        return token if token else None

    @token.setter
    def token (self, token: str):
        token = token.encode() if token is not None else None
        status = get_fxnc().FXNConfigurationSetToken(self.__configuration, token)
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to set configuration token with error: {status_to_error(status)}")            

    @property
    def acceleration (self) -> Acceleration:
        acceleration = c_int()
        status = get_fxnc().FXNConfigurationGetAcceleration(
            self.__configuration,
            byref(acceleration)
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get configuration acceleration with error: {status_to_error(status)}")
        return self.__to_acceleration_str(acceleration.value)

    @acceleration.setter
    def acceleration (self, acceleration: Acceleration):
        status = get_fxnc().FXNConfigurationSetAcceleration(
            self.__configuration,
            self.__to_acceleration_int(acceleration)
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to set configuration acceleration with error: {status_to_error(status)}")

    @property
    def device (self):
        device = c_void_p()
        status = get_fxnc().FXNConfigurationGetDevice(
            self.__configuration,
            byref(device)
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to get configuration device with error: {status_to_error(status)}")
        return device if device.value else None            

    @device.setter
    def device (self, device):
        status = get_fxnc().FXNConfigurationSetDevice(self.__configuration, device)
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to set configuration device with error: {status_to_error(status)}")

    def add_resource (self, type: str, path: Path):
        status = get_fxnc().FXNConfigurationAddResource(
            self.__configuration,
            type.encode(),
            str(path).encode()
        )
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to add configuration resource with error: {status_to_error(status)}")

    @classmethod
    def get_unique_id (cls) -> str:
        buffer = create_string_buffer(2048)
        status = get_fxnc().FXNConfigurationGetUniqueID(buffer, len(buffer))
        if status != FXNStatus.OK:
            raise RuntimeError(f"Failed to retrieve configuration identifier with error: {status_to_error(status)}")
        return buffer.value.decode("utf-8")

    @classmethod
    def get_client_id (cls) -> str:
        buffer = create_string_buffer(64)
        status = get_fxnc().FXNConfigurationGetClientID(buffer, len(buffer))
        if status == FXNStatus.OK:
            return buffer.value.decode("utf-8")
        else:
            raise RuntimeError(f"Failed to retrieve client identifier with error: {status_to_error(status)}")
        
    def __enter__ (self):
        return self
    
    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__configuration:
            get_fxnc().FXNConfigurationRelease(self.__configuration)
        self.__configuration = None

    def __to_acceleration_int (self, value: Acceleration) -> int:
        match value:
            case "auto": return 0
            case "cpu": return 1
            case "gpu": return 2
            case "npu": return 4

    def __to_acceleration_str (self, value: int) -> Acceleration:
        match value:
            case 0: return "auto"
            case 1: return "cpu"
            case 2: return "gpu"
            case 4: return "npu"
            case _: return None


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from asyncio import run as run_async
from io import BytesIO
from numpy import array_repr, ndarray
from pathlib import Path, PurePath
from PIL import Image
from rich import print_json
from tempfile import mkstemp
from typer import Argument, Context, Option

from ..function import Function
from ..logging import CustomProgress, CustomProgressTask
from ..types import Prediction
from .auth import get_access_key

def create_prediction (
    tag: str=Argument(..., help="Predictor tag."),
    quiet: bool=Option(False, "--quiet", help="Suppress verbose logging when creating the prediction."),
    context: Context = 0
):
    run_async(_predict_async(tag, quiet=quiet, context=context))

async def _predict_async (tag: str, quiet: bool, context: Context):
    # Preload
    with CustomProgress(transient=True, disable=quiet):
        fxn = Function(get_access_key())
        with CustomProgressTask(
            loading_text="Preloading predictor...",
            done_text="Preloaded predictor"
        ):
            fxn.predictions.create(tag, inputs={ })
        with CustomProgressTask(loading_text="Making prediction..."):
            inputs = { }
            for i in range(0, len(context.args), 2):
                name = context.args[i].replace("-", "")
                value = _parse_value(context.args[i+1])
                inputs[name] = value
            prediction = fxn.predictions.create(tag, inputs=inputs)
    _log_prediction(prediction)

def _parse_value (value: str):
    """
    Parse a value from a CLI argument.

    Parameters:
        value (str): CLI input argument.

    Returns:
        bool | int | float | str | Path: Parsed value.
    """
    # Boolean
    if value == "true":
        return True
    if value == "false":
        return False
    # Integer
    try:
        return int(value)
    except ValueError:
        pass
    # Float
    try:
        return float(value)
    except ValueError:
        pass
    # File
    if value.startswith("@"):
        path = Path(value[1:]).expanduser().resolve()
        if path.suffix in [".txt", ".md"]:
            with open(path) as f:
                return f.read()
        elif path.suffix in [".jpg", ".png"]:
            return Image.open(path)
        else:
            with open(path, "rb") as f:
                return BytesIO(f.read())
    # String
    return value

def _log_prediction (prediction: Prediction):
    images = [value for value in prediction.results or [] if isinstance(value, Image.Image)]
    prediction.results = [_serialize_value(value) for value in prediction.results] if prediction.results is not None else None
    print_json(data=prediction.model_dump())
    for image in images:
        image.show()

def _serialize_value (value):
    if isinstance(value, ndarray):
        return array_repr(value)
    if isinstance(value, Image.Image):
        _, path = mkstemp(suffix=".png" if value.mode == "RGBA" else ".jpg")
        value.save(path)
        return path
    if isinstance(value, BytesIO):
        return str(value)
    if isinstance(value, PurePath):
        return str(value)
    return value


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from .metadata import (
    CoreMLInferenceMetadata, GGUFInferenceMetadata,
    ONNXInferenceMetadata, ONNXRuntimeInferenceSessionMetadata
)
from .remote import RemoteAcceleration


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from contextvars import ContextVar
from rich.console import Console, ConsoleOptions, RenderResult
from rich.progress import Progress, ProgressColumn, SpinnerColumn, TextColumn
from rich.text import Text
from rich.traceback import Traceback
from types import MethodType

current_progress = ContextVar("current_progress", default=None)
progress_task_stack = ContextVar("progress_task_stack", default=[])

class CustomSpinnerColumn (SpinnerColumn):
    
    def __init__ (
        self,
        spinner_name="dots",
        success_text="[bold green]✔[/bold green]",
        failure_text="[bright_red]✘[/bright_red]",
        style="",
    ):
        super().__init__(spinner_name=spinner_name, style=style)
        self.success_text = success_text
        self.failure_text = failure_text

    def render (self, task):
        done_text = (
            self.failure_text
            if task.fields.get("status") == "error"
            else self.success_text
        )
        return done_text if task.finished else self.spinner

class CustomTextColumn (TextColumn):
    """Custom text column that changes color based on task status"""
    
    def __init__ (self, text_format="{task.description}"):
        super().__init__(text_format)
    
    def render (self, task):
        # Indent and color
        description = task.description
        indent_level = task.fields.get("indent_level", 0)
        indent = self.__get_indent(indent_level)
        task.description = f"{indent}{description}"
        if task.fields.get("status") == "error":
            task.description = f"[bright_red]{task.description}[/bright_red]"
        # Render
        text = super().render(task)
        task.description = description        
        # Return
        return text
    
    def __get_indent (self, level: int) -> str:
        if level == 0:
            return ""
        indicator = "└── "
        return " " * len(indicator) * (level - 1) + indicator

class CustomProgress(Progress):

    def __init__ (
        self,
        *columns: ProgressColumn,
        console=None,
        auto_refresh=True,
        refresh_per_second = 10,
        speed_estimate_period=30,
        transient=False,
        redirect_stdout=True,
        redirect_stderr=True,
        get_time=None,
        disable=False,
        expand=False
    ):
        default_columns = list(columns) if len(columns) > 0 else [
            CustomSpinnerColumn(),
            CustomTextColumn("[progress.description]{task.description}"),
        ]
        super().__init__(
            *default_columns,
            console=console,
            auto_refresh=auto_refresh,
            refresh_per_second=refresh_per_second,
            speed_estimate_period=speed_estimate_period,
            transient=transient,
            redirect_stdout=redirect_stdout,
            redirect_stderr=redirect_stderr,
            get_time=get_time,
            disable=disable,
            expand=expand
        )
        self.default_columns = default_columns

    def __enter__ (self):
        self._token = current_progress.set(self)
        self._stack_token = progress_task_stack.set([])
        return super().__enter__()

    def __exit__ (self, exc_type, exc_val, exc_tb):
        current_progress.reset(self._token)
        progress_task_stack.reset(self._stack_token)
        return super().__exit__(exc_type, exc_val, exc_tb)
    
    def get_renderables (self):
        for task in self.tasks:
            task_columns = task.fields.get("columns") or list()
            self.columns = self.default_columns + task_columns
            yield self.make_tasks_table([task])

class CustomProgressTask:

    def __init__ (
        self,
        *,
        loading_text: str,
        done_text: str=None,
        columns: list[ProgressColumn]=None
    ):
        self.loading_text = loading_text
        self.done_text = done_text
        self.task_id = None
        self.columns = columns

    def __enter__ (self):
        progress = current_progress.get()
        if progress is not None:
            self.task_id = progress.add_task(
                self.loading_text,
                total=1,
                columns=self.columns,
                indent_level=len(progress_task_stack.get())
            )
            current_stack = progress_task_stack.get()
            progress_task_stack.set(current_stack + [self.task_id])
        return self

    def __exit__ (self, exc_type, exc_val, exc_tb):
        progress = current_progress.get()
        if progress is not None and self.task_id is not None:
            current_task = progress._tasks[self.task_id]
            progress.update(
                self.task_id,
                description=self.done_text or current_task.description,
                completed=current_task.total,
                status="error" if exc_type is not None else current_task.fields.get("status")
            )
            current_stack = progress_task_stack.get()
            if current_stack:
                progress_task_stack.set(current_stack[:-1])
        self.task_id = None
        return False

    def update (self, **kwargs):
        progress = current_progress.get()
        if progress is None or self.task_id is None:
            return
        progress.update(self.task_id, **kwargs)

    def finish (self, message: str):
        self.done_text = message
    
class TracebackMarkupConsole (Console):

    def print(
        self,
        *objects,
        sep = " ",
        end = "\n",
        style = None,
        justify = None,
        overflow = None,
        no_wrap = None,
        emoji = None,
        markup = None,
        highlight = None,
        width = None,
        height = None,
        crop = True,
        soft_wrap = None,
        new_line_start = False
    ):
        traceback = objects[0]
        if isinstance(traceback, Traceback):
            stack = traceback.trace.stacks[0]
            original_rich_console = traceback.__rich_console__
            def __rich_console__ (self: Traceback, console: Console, options: ConsoleOptions) -> RenderResult:
                for renderable in original_rich_console(console, options):
                    if (
                        isinstance(renderable, Text) and
                        any(part.startswith(f"{stack.exc_type}:") for part in renderable._text)
                    ):
                        yield Text.assemble(
                            (f"{stack.exc_type}: ", "traceback.exc_type"),
                            Text.from_markup(stack.exc_value)
                        )
                    else:
                        yield renderable
            traceback.__rich_console__ = MethodType(__rich_console__, traceback)
        return super().print(
            *objects,
            sep=sep,
            end=end,
            style=style,
            justify=justify,
            overflow=overflow,
            no_wrap=no_wrap,
            emoji=emoji,
            markup=markup,
            highlight=highlight,
            width=width,
            height=height,
            crop=crop,
            soft_wrap=soft_wrap,
            new_line_start=new_line_start
        )


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ..client import FunctionClient, FunctionAPIError
from ..types import User

class UserService:

    def __init__ (self, client: FunctionClient) -> None:
        self.client = client

    def retrieve (self) -> User:
        """
        Retrieve the current user.

        Returns:
            User: User.
        """
        try:
            return self.client.request(
                method="GET",
                path="/users",
                response_type=User
            )
        except FunctionAPIError as error:
            if error.status_code == 401:
                return None
            raise


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from abc import ABC, abstractmethod
from hashlib import sha256
from pathlib import Path
from pydantic import BaseModel
from requests import put
from rich.progress import BarColumn, TextColumn
from typing import Literal

from .function import Function
from .logging import CustomProgressTask

class WorkdirCommand (BaseModel):
    kind: Literal["workdir"] = "workdir"
    path: str

class EnvCommand (BaseModel):
    kind: Literal["env"] = "env"
    env: dict[str, str]

class UploadableCommand (BaseModel, ABC):
    from_path: str
    to_path: str
    manifest: dict[str, str] | None = None

    @abstractmethod
    def get_files (self) -> list[Path]:
        pass

class UploadFileCommand (UploadableCommand):
    kind: Literal["upload_file"] = "upload_file"
    
    def get_files (self) -> list[Path]:
        return [Path(self.from_path).resolve()]

class UploadDirectoryCommand (UploadableCommand):
    kind: Literal["upload_dir"] = "upload_dir"

    def get_files (self) -> list[Path]:
        from_path = Path(self.from_path)
        if not from_path.is_absolute():
            raise ValueError("Cannot upload directory because directory path must be absolute")
        return [file for file in from_path.rglob("*") if file.is_file()]
    
class EntrypointCommand (UploadableCommand):
    kind: Literal["entrypoint"] = "entrypoint"
    name: str

    def get_files (self) -> list[Path]:
        return [Path(self.from_path).resolve()]

class PipInstallCommand (BaseModel):
    kind: Literal["pip_install"] = "pip_install"
    packages: list[str]

class AptInstallCommand (BaseModel):
    kind: Literal["apt_install"] = "apt_install"
    packages: list[str]

Command = (
    WorkdirCommand          |
    EnvCommand              |
    UploadFileCommand       |
    UploadDirectoryCommand  |
    PipInstallCommand       |
    AptInstallCommand       |
    EntrypointCommand
)

class Sandbox (BaseModel):
    """
    Sandbox which defines a containerized environment for compiling your Python function.
    """
    commands: list[Command] = []

    def workdir (self, path: str | Path) -> Sandbox:
        """
        Change the current working directory for subsequent commands.

        Parameters:
            path (str | Path): Path to change to.
        """
        command = WorkdirCommand(path=str(path))
        return Sandbox(commands=self.commands + [command])

    def env (self, **env: str) -> Sandbox:
        """
        Set environment variables in the sandbox.
        """
        command = EnvCommand(env=env)
        return Sandbox(commands=self.commands + [command])

    def upload_file (
        self,
        from_path: str | Path,
        to_path: str | Path = "./"
    ) -> Sandbox:
        """
        Upload a file to the sandbox.

        Parameters:
            from_path (str | Path): File path on the local file system.
            to_path (str | Path): Remote path to upload file to.
        """
        from_path = from_path if isinstance(from_path, Path) else Path(from_path)
        command = UploadFileCommand(
            from_path=str(from_path.resolve()),
            to_path=str(to_path)
        )
        return Sandbox(commands=self.commands + [command])

    def upload_directory (
        self,
        from_path: str | Path,
        to_path: str | Path = "."
    ) -> Sandbox:
        """
        Upload a directory to the sandbox.

        Parameters:
            from_path (str | Path): Directory path on the local file system.
            to_path (str | Path): Remote path to upload directory to.
        """
        from_path = from_path if isinstance(from_path, Path) else Path(from_path)
        command = UploadDirectoryCommand(
            from_path=str(from_path.resolve()),
            to_path=str(to_path)
        )
        return Sandbox(commands=self.commands + [command])

    def pip_install (self, *packages: str) -> Sandbox:
        """
        Install Python packages in the sandbox.

        Parameters:
            packages (list): Packages to install.
        """
        command = PipInstallCommand(packages=packages)
        return Sandbox(commands=self.commands + [command])

    def apt_install (self, *packages: str) -> Sandbox:
        """
        Install Debian packages in the sandbox.

        Parameters:
            packages (list): Packages to install.
        """
        command = AptInstallCommand(packages=packages)
        return Sandbox(commands=self.commands + [command])

    def populate (self, fxn: Function=None) -> Sandbox: # CHECK # In place
        """
        Populate all metadata.
        """
        fxn = fxn if fxn is not None else Function()
        entrypoint = next(cmd for cmd in self.commands if isinstance(cmd, EntrypointCommand))
        entry_path = Path(entrypoint.from_path).resolve()
        for command in self.commands:
            if isinstance(command, UploadableCommand):
                cwd = Path.cwd()
                from_path = Path(command.from_path)
                to_path = Path(command.to_path)
                if not from_path.is_absolute():
                    from_path = (entry_path / from_path).resolve()
                    command.from_path = str(from_path)
                files = command.get_files()
                name = from_path.relative_to(cwd) if from_path.is_relative_to(cwd) else from_path.resolve()
                with CustomProgressTask(
                    loading_text=f"Uploading [light_slate_blue]{name}[/light_slate_blue]...",
                    done_text=f"Uploaded [light_slate_blue]{name}[/light_slate_blue]",
                    columns=[
                        BarColumn(),
                        TextColumn("{task.completed}/{task.total}")
                    ]
                ) as task:
                    manifest = { }
                    for idx, file in enumerate(files):
                        rel_file_path = file.relative_to(from_path) if from_path.is_dir() else file.name
                        dst_path = to_path / rel_file_path
                        checksum = self.__upload_file(file, fxn=fxn)
                        manifest[str(dst_path)] = checksum
                        task.update(total=len(files), completed=idx+1)
                    command.manifest = manifest
        return self

    def __upload_file (self, path: Path, fxn: Function) -> str:
        if not path.is_file():
            raise ValueError(f"Cannot upload file at path {path} because it is not a file")
        hash = self.__compute_hash(path)
        try:
            fxn.client.request(method="HEAD", path=f"/resources/{hash}")
        except:
            resource = fxn.client.request(
                method="POST",
                path="/resources",
                body={ "name": hash },
                response_type=_Resource
            )
            with path.open("rb") as f:
                put(resource.url, data=f).raise_for_status()
        return hash
    
    def __compute_hash (self, path: Path) -> str:
        hash = sha256()
        with path.open("rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash.update(chunk)
        return hash.hexdigest()

class _Resource (BaseModel):
    url: str


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from functools import wraps
from inspect import signature as get_signature, Signature
from typing import get_origin, Callable, Generator, Iterator, TypeVar

from ..client import FunctionClient
from ..services import PredictionService as EdgePredictionService
from ..types import Acceleration
from .prediction import PredictionService
from .remote import RemoteAcceleration

F = TypeVar("F", bound=Callable[..., object])

class BetaClient:
    """
    Client for incubating features.
    """
    predictions: PredictionService
    
    def __init__ (
        self,
        client: FunctionClient,
        *,
        predictions: EdgePredictionService
    ):
        self.predictions = PredictionService(client)
        self.__edge_predictions = predictions

    def predict ( # INCOMPLETE # Preload
        self,
        tag: str,
        *,
        remote: bool=False,
        acceleration: Acceleration | RemoteAcceleration="auto",
        preload: bool=True
    ) -> Callable[[F], F]:
        """
        Create a prediction and return results when the decorated function is invoked.

        Parameters:
            tag (str): Predictor tag.
            remote (bool): Whether to create the prediction remotely.
            acceleration (Acceleration | RemoteAcceleration): Prediction acceleration.
            preload (bool): Whether to preload the predictor on the first run.
        """
        def decorator(func: F) -> F:
            signature = get_signature(func)
            @wraps(func)
            def wrapper(*args, **kwargs):
                bound_args = signature.bind(*args, **kwargs)
                bound_args.apply_defaults()
                stream = (
                    signature.return_annotation is not Signature.empty and
                    get_origin(signature.return_annotation) in [Iterator, Generator]
                )
                create_func = self.predictions.remote.create if remote else self.__edge_predictions.create
                def _predict (): # INCOMPLETE
                    prediction = create_func(
                        tag=tag,
                        inputs=bound_args.arguments,
                        acceleration=acceleration
                    )
                    if prediction.error:
                        raise RuntimeError(prediction.error)
                    return tuple(prediction.results) if len(prediction.results) > 1 else prediction.results[0]
                result = _predict()
                return result
            return wrapper
        return decorator


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

# https://github.com/fxnai/fxnc

from .configuration import Configuration
from .map import ValueMap
from .prediction import Prediction
from .predictor import Predictor
from .stream import PredictionStream
from .value import Value, ValueFlags


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from enum import Enum
from pydantic import AliasChoices, BaseModel, ConfigDict, Field
from typing import Any

from .dtype import Dtype
from .user import User

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Compiling = "COMPILING"
    Active = "ACTIVE"
    Invalid = "INVALID"
    Archived = "ARCHIVED"

class EnumerationMember (BaseModel):
    """
    Prediction parameter enumeration member.

    Members:
        name (str): Enumeration member name.
        value (str | int): Enumeration member value.
    """
    name: str = Field(description="Enumeration member name.")
    value: str | int = Field(description="Enumeration member value.")

class Parameter (BaseModel):
    """
    Predictor parameter.

    Members:
        name (str): Parameter name.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by Function.
        description (str): Parameter description.
        optional (bool): Whether the parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        enumeration (list): Parameter value choices for enumeration parameters.
        value_schema (dict): Parameter JSON schema. This is only populated for `list` and `dict` parameters.
    """
    name: str = Field(description="Parameter name.")
    type: Dtype | None = Field(default=None, description="Parameter type. This is `None` if the type is unknown or unsupported by Function.")
    description: str | None = Field(default=None, description="Parameter description.")
    optional: bool | None = Field(default=None, description="Whether the parameter is optional.")
    range: tuple[float, float] | None = Field(default=None, description="Parameter value range for numeric parameters.")
    enumeration: list[EnumerationMember] | None = Field(default=None, description="Parameter value choices for enumeration parameters.")
    value_schema: dict[str, Any] | None = Field(default=None, description="Parameter JSON schema. This is only populated for `list` and `dict` parameters.", serialization_alias="schema", validation_alias=AliasChoices("schema", "value_schema"))
    model_config = ConfigDict(arbitrary_types_allowed=True)

class Signature (BaseModel):
    """
    Predictor signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: list[Parameter] = Field(description="Input parameters.")
    outputs: list[Parameter] = Field(description="Output parameters.")

class Predictor (BaseModel):
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (User): Predictor owner.
        name (str): Predictor name.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        signature (Signature): Predictor signature.
        created (str): Date created.
        description (str): Predictor description.
        card (str): Predictor card.
        media (str): Predictor media URL.
        license (str): Predictor license URL.
    """
    tag: str = Field(description="Predictor tag.")
    owner: User = Field(description="Predictor owner.")
    name: str = Field(description="Predictor name.")
    status: PredictorStatus = Field(description="Predictor status.")
    access: AccessMode = Field(description="Predictor access.")
    signature: Signature = Field(description="Predictor signature.")
    created: str = Field(description="Date created.")
    description: str | None = Field(default=None, description="Predictor description.")
    card: str | None = Field(default=None, description="Predictor card.")
    media: str | None = Field(default=None, description="Predictor media URL.")
    license: str | None = Field(default=None, description="Predictor license URL.")


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import byref, c_void_p
from typing import final

from .fxnc import get_fxnc, status_to_error, FXNStatus
from .prediction import Prediction

@final
class PredictionStream:

    def __init__ (self, stream):
        self.__stream = stream

    def __iter__ (self):
        return self
    
    def __next__ (self) -> Prediction:
        prediction = c_void_p()
        status = get_fxnc().FXNPredictionStreamReadNext(self.__stream, byref(prediction))
        if status == FXNStatus.ERROR_INVALID_OPERATION:
            raise StopIteration()
        elif status != FXNStatus.OK:
            raise RuntimeError(f"Failed to read next prediction in stream with error: {status_to_error(status)}")
        else:
            return Prediction(prediction)

    def __enter__ (self):
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__stream:
            get_fxnc().FXNPredictionStreamRelease(self.__stream)
        self.__stream = None


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from .dtype import Dtype
from .prediction import Acceleration, Prediction, PredictionResource
from .predictor import AccessMode, EnumerationMember, Parameter, Predictor, PredictorStatus, Signature
from .user import User


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from datetime import datetime
from pathlib import Path
from pydantic import BaseModel
from rich import print_json
from typer import Argument, Option
from typing_extensions import Annotated

from ..function import Function
from ..logging import CustomProgress, CustomProgressTask
from .auth import get_access_key

def retrieve_source (
    predictor: Annotated[str, Option(help="Predictor tag.")] = None,
    prediction: Annotated[str, Option(help="Prediction identifier. If specified, this MUST be from a prediction returned by the Function API.")] = None,
    output: Annotated[Path, Option(help="Path to output source file.")] = Path("predictor.cpp")
):
    if not ((predictor is not None) ^ (prediction is not None)):
        raise ValueError(f"Predictor tag or prediction identifier must be provided, but not both.")
    fxn = Function(get_access_key())
    with CustomProgress(transient=True):
        if prediction is None:
            with CustomProgressTask(loading_text="Creating prediction..."):
                empty_prediction = fxn.predictions.create(tag=predictor)
                prediction = empty_prediction.id
        with CustomProgressTask(loading_text="Retrieving source..."):
            source = fxn.client.request(
                method="GET",
                path=f"/predictions/{prediction}/source",
                response_type=_PredictionSource
            )
            output.write_text(source.code)
            source.code = str(output.resolve())
            print_json(data=source.model_dump(mode="json", by_alias=True))

class _PredictionSource (BaseModel):
    tag: str
    target: str
    code: str
    created: datetime
    compiled: datetime
    latency: float # millis


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from ctypes import byref, c_void_p
from typing import final

from .configuration import Configuration
from .fxnc import get_fxnc, status_to_error, FXNStatus
from .map import ValueMap
from .prediction import Prediction
from .stream import PredictionStream

@final
class Predictor:

    def __init__ (self, configuration: Configuration):
        predictor = c_void_p()
        status = get_fxnc().FXNPredictorCreate(configuration._Configuration__configuration, byref(predictor))
        if status == FXNStatus.OK:
            self.__predictor = predictor
        else:
            raise RuntimeError(f"Failed to create predictor with error: {status_to_error(status)}")

    def create_prediction (self, inputs: ValueMap) -> Prediction:
        prediction = c_void_p()
        status = get_fxnc().FXNPredictorCreatePrediction(
            self.__predictor,
            inputs._ValueMap__map,
            byref(prediction)
        )
        if status == FXNStatus.OK:
            return Prediction(prediction)
        else:
            raise RuntimeError(f"Failed to create prediction with error: {status_to_error(status)}")

    def stream_prediction (self, inputs: ValueMap) -> PredictionStream:
        stream = c_void_p()
        status = get_fxnc().FXNPredictorStreamPrediction(
            self.__predictor,
            inputs._ValueMap__map,
            byref(stream)
        )
        if status == FXNStatus.OK:
            return PredictionStream(stream)
        else:
            raise RuntimeError(f"Failed to stream prediction with error: {status_to_error(status)}")

    def __enter__ (self):
        return self

    def __exit__ (self, exc_type, exc_value, traceback):
        self.__release()

    def __release (self):
        if self.__predictor:
            get_fxnc().FXNPredictorRelease(self.__predictor)
        self.__predictor = None


#
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict, is_dataclass
from datetime import datetime, timezone
from io import BytesIO
from numpy import array, ndarray
from pathlib import Path
from PIL import Image
from pydantic import BaseModel
from requests import get
from rich.progress import BarColumn, DownloadColumn, TransferSpeedColumn, TimeRemainingColumn
from tempfile import gettempdir, NamedTemporaryFile
from typing import Iterator
from urllib.parse import urlparse

from ..c import Configuration, Predictor, Prediction as CPrediction, Value as CValue, ValueFlags, ValueMap
from ..client import FunctionClient
from ..logging import CustomProgressTask
from ..types import Acceleration, Prediction, PredictionResource

Value = (
    None                |
    float               |
    int                 |
    bool                |
    ndarray             |
    str                 |
    list[object]        |
    dict[str, object]   |
    Image.Image         |
    BytesIO             |
    memoryview
)

class PredictionService:

    def __init__ (self, client: FunctionClient):
        self.client = client
        self.__cache = { }
        self.__cache_dir = self.__class__.__get_home_dir() / ".fxn" / "cache"
        self.__cache_dir.mkdir(parents=True, exist_ok=True)

    def ready (self, tag: str, **kwargs) -> bool:
        """
        Check whether a predictor has been preloaded and is ready to make predictions.

        Parameters:
            tag (str): Predictor tag.

        Returns:
            bool: Whether the predictor is ready to make predictions.
        """
        return tag in self.__cache

    def create (
        self,
        tag: str,
        *,
        inputs: dict[str, Value] | None=None,
        acceleration: Acceleration="auto",
        device=None,
        client_id: str=None,
        configuration_id: str=None
    ) -> Prediction:
        """
        Create a prediction.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values.
            acceleration (Acceleration): Prediction acceleration.
            client_id (str): Function client identifier. Specify this to override the current client identifier.
            configuration_id (str): Configuration identifier. Specify this to override the current client configuration identifier.

        Returns:
            Prediction: Created prediction.
        """
        if inputs is None:
            return self.__create_raw_prediction(
                tag=tag,
                client_id=client_id,
                configuration_id=configuration_id
            )
        predictor = self.__get_predictor(
            tag=tag,
            acceleration=acceleration,
            device=device,
            client_id=client_id,
            configuration_id=configuration_id
        )
        with (
            self.__to_value_map(inputs) as input_map,
            predictor.create_prediction(input_map) as prediction
        ):
            return self.__to_prediction(tag, prediction)

    def stream (
        self,
        tag: str,
        *,
        inputs: dict[str, Value],
        acceleration: Acceleration="auto",
        device=None
    ) -> Iterator[Prediction]:
        """
        Stream a prediction.

        Parameters:
            tag (str): Predictor tag.
            inputs (dict): Input values.
            acceleration (Acceleration): Prediction acceleration.

        Returns:
            Prediction: Created prediction.
        """
        predictor = self.__get_predictor(
            tag=tag,
            acceleration=acceleration,
            device=device,
        )
        with (
            self.__to_value_map(inputs) as input_map,
            predictor.stream_prediction(input_map) as stream
        ):
            for prediction in stream:
                with prediction:
                    yield self.__to_prediction(tag, prediction)

    def __create_raw_prediction (
        self,
        tag: str,
        client_id: str=None,
        configuration_id: str=None
    ) -> Prediction:
        client_id = client_id if client_id is not None else Configuration.get_client_id()
        configuration_id = configuration_id if configuration_id is not None else Configuration.get_unique_id()
        prediction = self.client.request(
            method="POST",
            path="/predictions",
            body={
                "tag": tag,
                "clientId": client_id,
                "configurationId": configuration_id,
            },
            response_type=Prediction
        )
        return prediction

    def __get_predictor (
        self,
        tag: str,
        acceleration: Acceleration="auto",
        device=None,
        client_id: str=None,
        configuration_id: str=None
    ) -> Predictor:
        if tag in self.__cache:
            return self.__cache[tag]
        prediction = self.__create_raw_prediction(
            tag=tag,
            client_id=client_id,
            configuration_id=configuration_id
        )
        with Configuration() as configuration:
            configuration.tag = prediction.tag
            configuration.token = prediction.configuration
            configuration.acceleration = acceleration
            configuration.device = device
            for resource in prediction.resources:
                path = self.__download_resource(resource)
                configuration.add_resource(resource.type, path)
            predictor = Predictor(configuration)
        self.__cache[tag] = predictor
        return predictor
    
    def __to_value_map (self, inputs: dict[str, Value]) -> ValueMap:
        map = ValueMap()
        for name, value in inputs.items():
            map[name] = self.__to_value(value)
        return map

    def __to_value (
        self,
        value: Value,
        *,
        flags: ValueFlags=ValueFlags.NONE
    ) -> CValue:
        value = self.__class__.__try_ensure_serializable(value)
        if value is None:
            return CValue.create_null()
        elif isinstance(value, bool):
            return self.__to_value(array(value, dtype="bool"), flags=flags | ValueFlags.COPY_DATA)
        elif isinstance(value, int):
            return self.__to_value(array(value, dtype="int32"), flags=flags | ValueFlags.COPY_DATA)
        elif isinstance(value, float):
            return self.__to_value(array(value, dtype="float32"), flags=flags | ValueFlags.COPY_DATA)
        elif isinstance(value, ndarray):
            return CValue.create_array(value, flags=flags)
        elif isinstance(value, str):
            return CValue.create_string(value)        
        elif isinstance(value, list):
            return CValue.create_list(value)
        elif isinstance(value, dict):
            return CValue.create_dict(value)
        elif isinstance(value, Image.Image):
            return CValue.create_image(value)
        elif isinstance(value, (bytes, bytearray, memoryview, BytesIO)):
            flags |= ValueFlags.COPY_DATA if not isinstance(value, memoryview) else 0
            view_or_bytes = value.getvalue() if isinstance(value, BytesIO) else value
            view = memoryview(view_or_bytes) if not isinstance(view_or_bytes, memoryview) else view_or_bytes
            return CValue.create_binary(view, flags=flags)            
        else:
            raise RuntimeError(f"Failed to convert object to Function value because object has an unsupported type: {type(value)}")

    def __to_prediction (self, tag: str, raw_prediction: CPrediction) -> Prediction:
        output_map = raw_prediction.results
        results = [output_map[output_map.key(idx)].to_object() for idx in range(len(output_map))] if output_map else None
        prediction = Prediction(
            id=raw_prediction.id,
            tag=tag,
            results=results,
            latency=raw_prediction.latency,
            error=raw_prediction.error,
            logs=raw_prediction.logs,
            created=datetime.now(timezone.utc).isoformat()
        )
        return prediction
    
    def __download_resource (self, resource: PredictionResource) -> Path:
        path = self.__get_resource_path(resource)
        if path.exists():
            return path
        path.parent.mkdir(parents=True, exist_ok=True)
        response = get(resource.url, stream=True)
        response.raise_for_status()
        size = int(response.headers.get("content-length", 0))
        stem = Path(urlparse(resource.url).path).name
        completed = 0
        color = "dark_orange" if not resource.type == "dso" else "purple"
        with (
            CustomProgressTask(
                loading_text=f"[{color}]{stem}[/{color}]",
                columns=[
                    BarColumn(),
                    DownloadColumn(),
                    TransferSpeedColumn(),
                    TimeRemainingColumn()
                ]
            ) as task,
            NamedTemporaryFile(mode="wb", delete=False) as tmp_file
        ):
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    tmp_file.write(chunk)
                    completed += len(chunk)
                    task.update(total=size, completed=completed)
            Path(tmp_file.name).replace(path)
        return path

    def __get_resource_path (self, resource: PredictionResource) -> Path:
        stem = Path(urlparse(resource.url).path).name
        path = self.__cache_dir / stem
        path = path / resource.name if resource.name else path
        return path

    @classmethod
    def __get_home_dir (cls) -> Path:
        try:
            check = Path.home() / ".fxntest"
            with open(check, "w") as f:
                f.write("fxn")
            check.unlink()
            return Path.home()
        except:
            return Path(gettempdir())

    @classmethod
    def __try_ensure_serializable (cls, object: object) -> object:
        if object is None:
            return object
        if isinstance(object, list):
            return [cls.__try_ensure_serializable(x) for x in object]
        if is_dataclass(object) and not isinstance(object, type):
            return asdict(object)
        if isinstance(object, BaseModel):
            return object.model_dump(mode="json", by_alias=True)
        return object


# 
#   Function
#   Copyright © 2025 NatML Inc. All Rights Reserved.
#

from collections.abc import Callable
from functools import wraps
from inspect import isasyncgenfunction, iscoroutinefunction
from pathlib import Path
from pydantic import BaseModel, ConfigDict, Field
from types import ModuleType
from typing import Literal

from .beta import (
    CoreMLInferenceMetadata, GGUFInferenceMetadata,
    ONNXInferenceMetadata, ONNXRuntimeInferenceSessionMetadata
)
from .sandbox import Sandbox
from .types import AccessMode

CompileTarget = Literal["android", "ios", "linux", "macos", "visionos", "wasm", "windows"]

CompileMetadata = (
    CoreMLInferenceMetadata             |
    GGUFInferenceMetadata               |
    ONNXInferenceMetadata               |
    ONNXRuntimeInferenceSessionMetadata
)

class PredictorSpec (BaseModel):
    """
    Descriptor of a predictor to be compiled.
    """
    tag: str = Field(description="Predictor tag.")
    description: str = Field(description="Predictor description. MUST be less than 100 characters long.", min_length=4, max_length=100)
    sandbox: Sandbox = Field(description="Sandbox to compile the function.")
    targets: list[str] | None = Field(description="Targets to compile this predictor for. Pass `None` to compile for our default targets.")
    access: AccessMode = Field(description="Predictor access.")
    card: str | None = Field(default=None, description="Predictor card (markdown).")
    media: str | None = Field(default=None, description="Predictor media URL.")
    license: str | None = Field(default=None, description="Predictor license URL. This is required for public predictors.")
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow", frozen=True)

def compile (
    tag: str,
    *,
    description: str,
    sandbox: Sandbox=None,
    trace_modules: list[ModuleType]=[],
    targets: list[CompileTarget]=None,
    access: AccessMode=AccessMode.Private,
    metadata: list[CompileMetadata]=[],
    card: str | Path=None,
    media: Path=None,
    license: str=None,
    **kwargs
):
    """
    Create a predictor by compiling a stateless function.

    Parameters:
        tag (str): Predictor tag.
        description (str): Predictor description. MUST be less than 100 characters long.
        sandbox (Sandbox): Sandbox to compile the function.
        trace_modules (list): Modules to trace and compile.
        targets (list): Targets to compile this predictor for. Pass `None` to compile for our default targets.
        access (AccessMode): Predictor access.
        metadata (list): Metadata to use while compiling the function.
        card (str | Path): Predictor card markdown string or path to card.
        media (Path): Predictor thumbnail image (jpeg or png) path.
        license (str): Predictor license URL. This is required for public predictors.
    """
    def decorator (func: Callable):
        # Check type
        if not callable(func):
            raise TypeError("Cannot compile non-function objects")
        if isasyncgenfunction(func) or iscoroutinefunction(func):
            raise TypeError(f"Entrypoint function '{func.__name__}' must be a regular function or generator")            
        # Gather metadata
        spec = PredictorSpec(
            tag=tag,
            description=description,
            sandbox=sandbox if sandbox is not None else Sandbox(),
            targets=targets,
            access=access,
            card=card.read_text() if isinstance(card, Path) else card,
            media=None, # INCOMPLETE
            license=license,
            trace_modules=trace_modules,
            metadata=metadata,
            **kwargs
        )
        # Wrap
        @wraps(func)
        def wrapper (*args, **kwargs):
            return func(*args, **kwargs)
        wrapper.__predictor_spec = spec
        return wrapper
    return decorator